/*
 * This file is part of family-shop-mgr.
 *
 * family-shop-mgr is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * family-shop-mgr is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with family-shop-mgr.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Author: Unai IRIGOYEN
 * Date: 12/07/2009
 *
 */

#include "ShoppingTreeItem.h"

/*******************************************************************/
ShoppingTreeItem::ShoppingTreeItem(const QVector<QVariant> &data,
                                   ShoppingTreeItem *parent) : m_itemType(NotDefined)
{
    parentItem = parent;
    itemData = data;
}

/*******************************************************************/
ShoppingTreeItem::~ShoppingTreeItem()
{
    qDeleteAll(childItems);
}

/*******************************************************************/
ShoppingTreeItem *ShoppingTreeItem::child(int number)
{
    return childItems.value(number);
}

/*******************************************************************/
int ShoppingTreeItem::childCount() const
{
    return childItems.count();
}

/*******************************************************************/
int ShoppingTreeItem::childNumber() const
{
    if (parentItem)
        return parentItem->childItems.indexOf(const_cast<ShoppingTreeItem*>(this));

    return 0;
}

/*******************************************************************/
int ShoppingTreeItem::columnCount() const
{
    return itemData.count();
}

/*******************************************************************/
QVariant ShoppingTreeItem::data(int column) const
{
    return itemData.value(column);
}

/*******************************************************************/
bool ShoppingTreeItem::insertChildren(int position, int count, int columns)
{
    if (position < 0 || position > childItems.size())
        return false;

    for (int row = 0; row < count; ++row) {
        QVector<QVariant> data(columns);
        ShoppingTreeItem *item = new ShoppingTreeItem(data, this);
        childItems.insert(position, item);
        emit childInserted(item);
    }

    return true;
}

/*******************************************************************/
bool ShoppingTreeItem::insertColumns(int position, int columns)
{
    if (position < 0 || position > itemData.size())
        return false;

    for (int column = 0; column < columns; ++column)
        itemData.insert(position, QVariant());

    foreach (ShoppingTreeItem *child, childItems)
        child->insertColumns(position, columns);

    return true;
}

/*******************************************************************/
ShoppingTreeItem *ShoppingTreeItem::parent()
{
    return parentItem;
}

/*******************************************************************/
bool ShoppingTreeItem::removeChildren(int position, int count)
{
    if (position < 0 || position + count > childItems.size())
        return false;

    for (int row = 0; row < count; ++row)
    {
        ShoppingTreeItem *item = childItems.takeAt(position);
        emit childRemoved(item);
        delete item;
    }

    return true;
}

/*******************************************************************/
bool ShoppingTreeItem::removeColumns(int position, int columns)
{
    if (position < 0 || position + columns > itemData.size())
        return false;

    for (int column = 0; column < columns; ++column)
        itemData.remove(position);

    foreach (ShoppingTreeItem *child, childItems)
        child->removeColumns(position, columns);

    return true;
}

/*******************************************************************/
bool ShoppingTreeItem::setData(int column, const QVariant &value)
{
    if (column < 0 || column >= itemData.size())
        return false;

    if(m_itemType == Category && column != 0)
        return false;

    itemData[column] = value;
    emit dataChanged(column);
    return true;
}

/*******************************************************************/
bool ShoppingTreeItem::setItemType(ItemType type)
{
    if(type == NotDefined)
        return false;

    m_itemType = type;

    // release enough resources in case we call wait more than once
    m_typeSemaphore.release(1000);

    return (m_itemType == type) ? true : false;
}

/*******************************************************************/
ShoppingTreeItem::ItemType ShoppingTreeItem::getItemType() const
{
    return m_itemType;
}

/*******************************************************************/
void ShoppingTreeItem::waitItemTypeDefinition()
{
    m_typeSemaphore.acquire();
}
