/*
 * This file is part of family-shop-mgr.
 *
 * family-shop-mgr is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * family-shop-mgr is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with family-shop-mgr.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Author: Unai IRIGOYEN
 * Date: 12/07/2009
 *
 */

#include "ShoppingTreeModel.h"

#include "ShoppingTreeItem.h"
#include <QFile>
#include <QApplication>
#include <QtGui>

/*******************************************************************/
ShoppingTreeModel::ShoppingTreeModel(const QString &xmlFileName,
                                     QObject *parent) :
QAbstractItemModel(parent), m_document("ShoppingList")
{
    QString error;
    int errLine;
    int errColumn;

    m_xmlFileName = QApplication::applicationDirPath() + "/" + xmlFileName;
    QFile file(m_xmlFileName);
    if(!file.open(QIODevice::ReadOnly))
        return;
    // Parse xml file
    if(!m_document.setContent(&file, true, &error, &errLine, &errColumn))
    {
        emit xmlParseError(error, errLine, errColumn);
        file.close();
        return;
    }
    file.close();

    QDomElement root = m_document.documentElement();
    if(root.tagName() != "shoppingList" || !root.hasAttribute("version"))
    {
        emit invalidDocument();
        return;
    }
    else if(root.attribute("version") == "1.0")
    {
        // set column titles
        QVector<QVariant> rootData;
        rootData << "Category/Item name"
                << "Quantity" << "Store";

        rootItem = new ShoppingTreeItem(rootData);
    }
    else
    {
        // upgrade document version if possible
        ;
    }

    QDomElement child = root.firstChildElement("category");
    while(!child.isNull())
    {
        // Parse all categories
        parseCategoryElement(child);
        child = child.nextSiblingElement("category");
    }

    child = root.firstChildElement("item");
    while(!child.isNull())
    {
        // parse all items which don't have category
        rootItem->insertChildren(
                rootItem->childCount(), 1,
                rootItem->columnCount());
        QVector<QVariant> columnData =
                getColumnsFromItemElement(child);
        rootItem->child(rootItem->childCount() - 1)->
                setItemType(ShoppingTreeItem::Item);
        for(int column = 0; column < columnData.size(); column++)
        {
            rootItem->child(rootItem->childCount() - 1)->setData(column, columnData[column]);
        }
        m_domElementForItem.insert(rootItem->child(rootItem->childCount() - 1),
                                   child);
        child = child.nextSiblingElement("item");
    }

    QHashIterator<ShoppingTreeItem*,QDomElement> i(m_domElementForItem);
    while(i.hasNext())
    {
        i.next();
        connect(i.key(), SIGNAL(childInserted(ShoppingTreeItem*)), this,
                SLOT(registerInsertedChild(ShoppingTreeItem*)));
        connect(i.key(), SIGNAL(childRemoved(ShoppingTreeItem*)), this,
                SLOT(deleteRemovedChild(ShoppingTreeItem*)));
    }
}

/*******************************************************************/
ShoppingTreeModel::~ShoppingTreeModel()
{
    delete rootItem;
}

/*******************************************************************/
bool ShoppingTreeModel::addCategory(QString name)
{
    QModelIndex parent = QModelIndex();

    if (!this->insertRow(parent.row()+1, parent))
        return false;

    for(int column = 0; column < this->columnCount(parent); ++column)
    {
        QModelIndex child = this->index(parent.row()+1, column, parent);
        ShoppingTreeItem* item = this->getItem(child);
        item->setItemType(ShoppingTreeItem::Category);
        this->setData(child, QVariant(name), Qt::EditRole);
    }
    return true;
}

/*******************************************************************/
bool ShoppingTreeModel::addSubCategory(QString name, int position,
                    const QModelIndex &parent)
{
    if (!this->insertRow(position, parent))
        return false;

    for(int column = 0; column < this->columnCount(parent); ++column)
    {
        QModelIndex child = this->index(parent.row()+1, column, parent);
        ShoppingTreeItem* item = this->getItem(child);
        item->setItemType(ShoppingTreeItem::Category);
        this->setData(child, QVariant(name), Qt::EditRole);
    }
    return true;
}

/*******************************************************************/
bool ShoppingTreeModel::addItem(QString name, int position,
             const QModelIndex &parent)
{
    if (!this->insertRow(position, parent))
        return false;

    for(int column = 0; column < this->columnCount(parent); ++column)
    {
        QModelIndex child = this->index(parent.row()+1, column, parent);
        ShoppingTreeItem* item = this->getItem(child);
        item->setItemType(ShoppingTreeItem::Item);
        this->setData(child, QVariant(name), Qt::EditRole);
    }
    return true;
}

/*******************************************************************/
bool ShoppingTreeModel::removeCategoryOrItem(const QModelIndex &index)
{
    return this->removeRow(index.row(), index.parent());
}

/*******************************************************************/
QVariant ShoppingTreeModel::data(const QModelIndex &index, int role) const
{
    if(!index.isValid())
        return QVariant();

    if(role != Qt::DisplayRole && role != Qt::EditRole)
        return QVariant();

    ShoppingTreeItem *item = getItem(index);
    return item->data(index.column());
}

/*******************************************************************/
Qt::ItemFlags ShoppingTreeModel::flags(const QModelIndex &index) const
{
    if(!index.isValid())
        return 0;

    return Qt::ItemIsEditable | Qt::ItemIsEnabled | Qt::ItemIsSelectable;
}

/*******************************************************************/
ShoppingTreeItem *ShoppingTreeModel::getItem(const QModelIndex &index) const
{
    if(index.isValid()){
        ShoppingTreeItem *item =
                static_cast<ShoppingTreeItem*>(index.internalPointer());
        if(item) return item;
    }

    return rootItem;
}

/*******************************************************************/
QVariant ShoppingTreeModel::headerData(int section, Qt::Orientation orientation,
                                       int role) const
{
    if(orientation == Qt::Horizontal && role == Qt::DisplayRole)
        return rootItem->data(section);

    return QVariant();
}

/*******************************************************************/
QModelIndex ShoppingTreeModel::index(int row, int column, const QModelIndex &parent) const
{
    if(parent.isValid() && parent.column() != 0)
        return QModelIndex();

    ShoppingTreeItem *parentItem = getItem(parent);

    ShoppingTreeItem *childItem = parentItem->child(row);
    if(childItem)
        return createIndex(row, column, childItem);
    else
        return QModelIndex();
}

/*******************************************************************/
bool ShoppingTreeModel::insertColumns(int position, int columns, const QModelIndex &parent)
{
    bool success;

    beginInsertColumns(parent, position, position + columns - 1);
    success = rootItem->insertColumns(position, columns);
    endInsertColumns();

    return success;
}

/*******************************************************************/
bool ShoppingTreeModel::insertRows(int position, int rows, const QModelIndex &parent)
{
    ShoppingTreeItem *parentItem = getItem(parent);
    bool success;

    beginInsertRows(parent, position, position + rows - 1);
    success = parentItem->insertChildren(position, rows, rootItem->columnCount());
    endInsertRows();

    return success;
}

/*******************************************************************/
QModelIndex ShoppingTreeModel::parent(const QModelIndex &index) const
{
    if(!index.isValid())
        return QModelIndex();

    ShoppingTreeItem *childItem = getItem(index);
    ShoppingTreeItem *parentItem = childItem->parent();

    if(parentItem == rootItem)
        return QModelIndex();

    return createIndex(parentItem->childNumber(), 0, parentItem);
}

/*******************************************************************/
bool ShoppingTreeModel::removeColumns(int position, int columns, const QModelIndex &parent)
{
    bool success;

    beginRemoveColumns(parent, position, position + columns - 1);
    success = rootItem->removeColumns(position, columns);
    endRemoveColumns();

    if (rootItem->columnCount() == 0)
        removeRows(0, rowCount());

    return success;
}

/*******************************************************************/
bool ShoppingTreeModel::removeRows(int position, int rows, const QModelIndex &parent)
{
    ShoppingTreeItem *parentItem = getItem(parent);
    bool success;

    beginRemoveRows(parent, position, position + rows - 1);
    success = parentItem->removeChildren(position, rows);
    endRemoveRows();

    return success;
}

/*******************************************************************/
int ShoppingTreeModel::rowCount(const QModelIndex &parent) const
{
    ShoppingTreeItem *parentItem = getItem(parent);

    return parentItem->childCount();
}

/*******************************************************************/
int ShoppingTreeModel::columnCount(const QModelIndex &parent) const
{
    return rootItem->columnCount();
}

/*******************************************************************/
bool ShoppingTreeModel::setData(const QModelIndex &index, const QVariant &value, int role)
{
    if(role != Qt::EditRole)
        return false;

    ShoppingTreeItem *item = getItem(index);

    // only "items" have more than one editable column
    if(index.column() != 0 && m_domElementForItem.value(item).tagName() != "item")
        return false;

    // edit item
    bool result = (item->setData(index.column(),value) &&
                   updateDomElement(item, index.column()));

    if(result)
        emit dataChanged(index, index);

    return result;
}

/*******************************************************************/
bool ShoppingTreeModel::setHeaderData(int section, Qt::Orientation orientation,
                                      const QVariant &value, int role)
{
    if(role != Qt::EditRole || orientation != Qt::Horizontal)
        return false;

    bool result = rootItem->setData(section, value);

    if(result)
        emit headerDataChanged(orientation, section, section);

    return result;
}

/*******************************************************************/
void ShoppingTreeModel::registerInsertedChild(ShoppingTreeItem *item)
{
    // wait until item type is defined
    item->waitItemTypeDefinition();

    QDomElement parentElement = m_domElementForItem.value(item->parent());
    QDomElement element;
    if(item->getItemType() == ShoppingTreeItem::Category)
        element = m_document.createElement("category");
    else if(item->getItemType() == ShoppingTreeItem::Item)
        element = m_document.createElement("item");
    else
        return;

    parentElement.appendChild(element);
    updateXmlFile();
    m_domElementForItem.insert(item, element);
    connect(item, SIGNAL(childInserted(ShoppingTreeItem*)), this,
            SLOT(registerInsertedChild(ShoppingTreeItem*)));
    connect(item, SIGNAL(childRemoved(ShoppingTreeItem*)), this,
            SLOT(deleteRemovedChild(ShoppingTreeItem*)));
}

/*******************************************************************/
void ShoppingTreeModel::deleteRemovedChild(ShoppingTreeItem *item)
{
    QDomElement element = m_domElementForItem.value(item);
    QDomNode parentNode = element.parentNode();
    parentNode.removeChild(element);
    updateXmlFile();
    m_domElementForItem.remove(item);
}

/*******************************************************************/
void ShoppingTreeModel::parseCategoryElement(const QDomElement &element,
                                             ShoppingTreeItem *parentItem)
{
    // if parent is null then add category to root
    if(!parentItem)
        parentItem = rootItem;

    ShoppingTreeItem *item;
    QDomElement child = element.firstChildElement("title");
    QString title = child.text();
    if(!title.isEmpty())
    {
        parentItem->insertChildren(parentItem->childCount(), 1,
                                   rootItem->columnCount());

        parentItem->child(parentItem->childCount() - 1)->
                setItemType(ShoppingTreeItem::Category);
        parentItem->child(parentItem->childCount() - 1)->setData(0, title);
        m_domElementForItem.insert(parentItem->child(parentItem->childCount() - 1),
                                   element);
        item = parentItem->child(parentItem->childCount() - 1);
    }
    else
    {
        emit invalidDocument();
        return;
    }

    // add each sub category and item to the tree
    child = child.nextSiblingElement();
    while(!child.isNull())
    {
        if(child.tagName() == "category")
        {
            parseCategoryElement(child, parentItem);
        }
        else if(child.tagName() == "item")
        {
            item->insertChildren(
                    item->childCount(), 1,
                    rootItem->columnCount());
            QVector<QVariant> columnData =
                    getColumnsFromItemElement(child);
            item->child(item->childCount() - 1)->setItemType(ShoppingTreeItem::Item);
            for(int column = 0; column < columnData.size(); column++)
            {
                item->child(item->childCount() - 1)->setData(column, columnData[column]);
            }
            m_domElementForItem.insert(item->child(item->childCount() - 1),
                                       child);
        }
        else
        {
            emit invalidDocument();
            return;
        }

        child = child.nextSiblingElement();
    }
}

/*******************************************************************/
QVector<QVariant> ShoppingTreeModel::getColumnsFromItemElement(const QDomElement &element)
{
    QVector<QVariant> data;
    QString title = element.firstChildElement("title").text();
    int quantity = element.firstChildElement("quantity").text().toInt();
    QString store = element.firstChildElement("store").text();
    if(title.isEmpty() || quantity < 0)
    {
        emit invalidDocument();
        return data;
    }

    data << title << quantity << store;
    return data;
}

/*******************************************************************/
bool ShoppingTreeModel::updateDomElement(ShoppingTreeItem *item, int column)
{
    QDomElement element = m_domElementForItem.value(item);

    if(element.isNull())
     return false;

    bool success;
    switch(column)/*******************************************************************/
    {
        case 0:
        {
            QDomElement oldTitleElement = element.firstChildElement("title");
            QDomElement newTitleElement = m_document.createElement("title");

            QDomText newTitleText = m_document.createTextNode(item->data(0).toString());
            newTitleElement.appendChild(newTitleText);

            element.replaceChild(newTitleElement, oldTitleElement);
            success = true;
            break;
        }
        case 1:
        {
            QDomElement oldQuantityElement = element.firstChildElement("quantity");
            QDomElement newQuantityElement = m_document.createElement("quantity");

            QDomText newQuantityText = m_document.createTextNode(item->data(1).toString());
            newQuantityElement.appendChild(newQuantityText);

            element.replaceChild(newQuantityElement, oldQuantityElement);
            success = true;
            break;
        }
        case 2:
        {
            QDomElement oldStoreElement = element.firstChildElement("store");
            QDomElement newStoreElement = m_document.createElement("store");

            QDomText newStoreText = m_document.createTextNode(item->data(0).toString());
            newStoreElement.appendChild(newStoreText);

            element.replaceChild(newStoreElement, oldStoreElement);
            success = true;
            break;
        }
        default:
            success = false;
    }

    updateXmlFile();

    return success;
}

/*******************************************************************/
void ShoppingTreeModel::updateXmlFile() const
{
    QFile xmlFile(m_xmlFileName);
    xmlFile.remove();
    xmlFile.open(QIODevice::WriteOnly);
    xmlFile.write(m_document.toByteArray(4));
    xmlFile.close();
}
